from .test_tool import TestTool


def gen_asset_relative_path(testtool: TestTool) -> str:
    """
    如果是本地上传测试工具压缩包到COS上，那么使用此生成规则
    """
    return (
        f"testtools/{testtool.lang}/{testtool.name}/archive/{testtool.version}.tar.gz"
    )



import tempfile
from pathlib import Path
from typing import Callable

from loguru import logger
from requests import HTTPError

from ..model.test_tool import (
    StableIndexMetaData,
    MetaDataHistory,
    TestTool,
    TestToolMetadata,
)
from ..service.generator import Generator
from ..service.testtool import sort_test_tools
from ..util.file import download_file_to


class MetaMerger:
    def __init__(self, testtool: TestTool, asset_url_gen: Callable[[TestTool], str]):
        self.testtool = testtool

        gen = Generator(self.testtool, asset_url_gen)
        self.metadata = gen.generate_meta_data()

    def merge_index_and_history(self, output_dir: Path) -> None:
        """
        合并新的索引文件和版本文件
        :param output_dir:  registry目录
        :return:
        """
        new_index = self._download_and_merge_stable_index()
        new_history = self._download_and_merge_meta_history()

        index_file = Path(output_dir) / "testtools" / "stable.index.json"
        index_file.parent.mkdir(exist_ok=True, parents=True)
        with open(index_file, "w", encoding="utf-8") as f:
            f.write(
                new_index.model_dump_json(by_alias=True, indent=2, exclude_none=True)
            )

        meta_file = (
            Path(output_dir)
            / "testtools"
            / self.testtool.lang
            / self.testtool.name
            / "metadata.json"
        )
        meta_file.parent.mkdir(exist_ok=True, parents=True)
        with open(
            meta_file,
            "w",
            encoding="utf-8",
        ) as f:
            f.write(
                new_history.model_dump_json(by_alias=True, indent=2, exclude_none=True)
            )

    def _download_and_merge_stable_index(self) -> StableIndexMetaData:
        """
        合并稳定索引文件内容

        一般用于新版本发布后索引文件的更新
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            stable_index_file = Path(tmpdir) / "stable_index.json"

            try:
                # 正常下载就进行合并操作
                download_file_to(
                    url=self.testtool.index_file, to_file=stable_index_file
                )
                return self._merge_stable_index(stable_index_file)
            except HTTPError as e:
                if e.response.status_code == 404:
                    # 没有索引文件，直接新建一个索引，不合并
                    return self._create_new_stable_index()
                else:
                    raise

    def _download_and_merge_meta_history(self) -> MetaDataHistory:
        """
        合并工具元数据的版本历史
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            metadata = Path(tmpdir) / "metadata.json"

            try:
                download_file_to(url=self.testtool.version_file, to_file=metadata)

                with open(metadata, "r", encoding="utf-8") as f:
                    old_history = MetaDataHistory.model_validate_json(f.read())
                    return self._merge_meta_history(old_history)
            except HTTPError as e:
                if e.response.status_code == 404:
                    return self._create_new_metadata_history()
                else:
                    raise

    def _create_new_stable_index(self) -> StableIndexMetaData:
        logger.info("Creating stable index...")
        meta = StableIndexMetaData(tools=[self.testtool])

        return meta

    def _merge_stable_index(self, stable_index: Path) -> StableIndexMetaData:
        logger.info(f"Merging {stable_index.name}")
        with open(stable_index, "r") as f:
            stable_result = StableIndexMetaData.model_validate_json(f.read())

            stable_result.merge_stable_index([self.testtool])
            stable_result.tools = sort_test_tools(stable_result.tools)

            logger.info(
                f"Merge stable index: {stable_result.model_dump_json(by_alias=True, indent=2, exclude_none=True)}"
            )
            return stable_result

    def _merge_meta_history(self, history: MetaDataHistory) -> MetaDataHistory:
        logger.info("Merging meta history...")
        if not history.versions:
            history.versions = []

        self._upsert_meta_history(tool_meta=self.metadata, history=history)

        # 在历史记录中增加一个固定的stable版本，方便用户使用
        stable_version = self.metadata.model_copy(deep=True)
        stable_version.meta.version = "stable"
        self._upsert_meta_history(tool_meta=stable_version, history=history)

        logger.info(
            f"Merge meta history result: {history.model_dump_json(by_alias=True, indent=2, exclude_none=True)}"
        )

        return history

    @staticmethod
    def _upsert_meta_history(
        tool_meta: TestToolMetadata, history: MetaDataHistory
    ) -> MetaDataHistory:
        for index, version in enumerate(history.versions):
            if version.meta.version == tool_meta.meta.version:
                history.versions[index] = tool_meta
                break
        else:
            history.versions.append(tool_meta)

        return history

    def _create_new_metadata_history(self) -> MetaDataHistory:
        # 读取本地生成好的metadata文件
        logger.info("Creating meta history...")
        history = MetaDataHistory(versions=[self.metadata])

        # 在历史记录中增加一个固定的stable版本，方便用户使用
        stable_version = self.metadata.model_copy(deep=True)
        stable_version.meta.version = "stable"
        history.versions.append(stable_version)

        return history



import sys
from pathlib import Path
from typing import Optional

import typer
from loguru import logger
from typing_extensions import Annotated

from .commands.meta_merger import MetaMerger
from .service.cos_sync import CosSyncService
from .service.pr_generator import PullRequestGenerator
from .service.testtool import get_testtool, github_asset_gen
from .service.validator import ToolValidator

app = typer.Typer()


@app.command()
def merge(tool_name: str, output: str, working_dir: Optional[str] = None) -> None:
    """
    合并工具版本元数据

    :param tool_name: 工具名称
    :param output: registry仓库本地目录
    :param working_dir: 可选工作目录
    """
    testtool = get_testtool(tool_name, working_dir)
    merger = MetaMerger(testtool, asset_url_gen=github_asset_gen)
    merger.merge_index_and_history(Path(output))


@app.command()
def pull_request(
    tool_name: Annotated[str, typer.Argument(help="工具名称")],
    working_dir: Annotated[Optional[str], typer.Argument(help="可选工作目录")] = None,
) -> None:
    """
    合并元数据之后，向项目提PR进行合并操作
    """
    testtool = get_testtool(tool_name, working_dir)
    pr_gen = PullRequestGenerator(testtool)
    pr_gen.merge_and_create_pull_request()


@app.command()
def validate(
    tool_name: str, working_dir: Optional[str] = None, verbose: bool = False
) -> None:
    """
    校验测试工具的testtools.yaml是否符合要求

    :param verbose: 是否显示详细日志
    :param tool_name: 工具名称
    :param working_dir: 可选工作目录
    """
    if not verbose:
        logger.remove()
        logger.add(sys.stdout, level="INFO")

    testtool = get_testtool(tool_name, working_dir)
    logger.info(f"✅ 测试工具 {testtool.name} 有效性校验通过")


@app.command()
def validate_json(working_dir: Optional[str] = None) -> None:
    """
    校验当前目录下的json文件是否符合要求

    :param working_dir: 可选工作目录
    """
    validator = ToolValidator(working_dir)
    validator.validate()


@app.command()
def sync_cos(working_dir: Optional[str] = None, force: bool = False) -> None:
    """
    同步数据到COS上

    :param working_dir: 可选工作目录
    :param force: 是否强制更新COS上的文件
    """
    sync = CosSyncService(working_dir)
    sync.sync_meta_data(force)


def cli_entry() -> None:
    app()


if __name__ == "__main__":
    app()



import hashlib
from pathlib import Path
from typing import Callable

from loguru import logger

from ..model.test_tool import (
    TestTool,
    TestToolTarget,
    TestToolMetadata,
)
from ..util.file import download_file_to


class Generator:
    def __init__(self, testtool: TestTool, asset_url_gen: Callable[[TestTool], str]):
        self.testtool = testtool
        self.asset_url_gen = asset_url_gen

    def generate_meta_data(self) -> TestToolMetadata:
        """
        生成测试工具元数据，包含工具信息和最新的版本信息
        """
        logger.info(
            f"Generating meta data for {self.testtool.model_dump_json(by_alias=True, indent=2, exclude_none=True)}"
        )

        if not self.testtool.legacy_spec:
            sha256 = self.compute_asset_sha256(self.testtool)
            metadata = TestToolMetadata(
                meta=self.testtool, target=self.generate_targets(self.testtool, sha256)
            )
            logger.info(
                f"Generated metadata: {metadata.model_dump_json(indent=2, by_alias=True, exclude_none=True)}"
            )
        else:
            logger.info(
                f"Testtool {self.testtool.name} is legacy tool, skip target generation."
            )
            metadata = TestToolMetadata(meta=self.testtool, target=[])

        return metadata

    def generate_targets(self, testtool: TestTool, sha256: str) -> list[TestToolTarget]:
        re: list[TestToolTarget] = []

        assert testtool.support_os
        assert testtool.support_arch

        for _os in testtool.support_os:
            for arch in testtool.support_arch:
                re.append(
                    TestToolTarget(
                        os=_os,
                        arch=arch,
                        downloadUrl=self.generate_asset_url(testtool),
                        sha256=sha256,
                    )
                )
        return re

    def generate_asset_url(self, testtool: TestTool) -> str:
        """
        指定产物的URL

        不同场景下，可能产物的URL需要定制，因此这个需要能动态配置
        """

        return self.asset_url_gen(testtool)

    def compute_asset_sha256(self, testtool: TestTool) -> str:
        # 读取本次发布的产物信息，并计算sha256值
        output_file = (
            Path("/tmp/testsolar/generate")
            / testtool.lang
            / testtool.name
            / "output.tar.gz"
        )
        output_file.parent.mkdir(parents=True, exist_ok=True)

        download_file_to(url=self.generate_asset_url(testtool), to_file=output_file)

        sha256_hash = hashlib.sha256()
        with open(output_file, "rb") as file:
            while True:
                data = file.read(65536)  # 一次读取64KB
                if not data:
                    break
                sha256_hash.update(data)

        sha256 = sha256_hash.hexdigest()

        logger.info(f"sha256: {sha256}")

        return sha256



from pathlib import Path
from typing import Union
import requests
from loguru import logger
from retry import retry


@retry(tries=5, delay=1, backoff=2)
def download_file_to(url: str, to_file: Union[str, Path]) -> None:
    logger.info(f"Downloading {url} to {to_file}")
    response = requests.get(url, stream=True)
    response.raise_for_status()

    with open(to_file, "wb") as f:
        # 逐块写入文件内容
        for chunk in response.iter_content(chunk_size=65536):
            if chunk:  # 过滤掉保持连接活动的新块
                f.write(chunk)

        logger.info(f"Downloaded {url} to {to_file}")
